############################################################################
#    Copyright (C) 2009 by GFZ Potsdam                                     #
#                                                                          #
#    author: Joachim Saul                                                  #
#    email:  saul@gfz-potsdam.de                                           #
#                                                                          #
#    This program is free software; you can redistribute it and/or modify  #
#    it under the terms of the GNU General Public License as published by  #
#    the Free Software Foundation; either version 2 of the License, or     #
#    (at your option) any later version.                                   #
#                                                                          #
#    This program is distributed in the hope that it will be useful,       #
#    but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#    GNU General Public License for more details.                          #
#                                                                          #
#    You should have received a copy of the GNU General Public License     #
#    along with this program; if not, write to the                         #
#    Free Software Foundation, Inc.,                                       #
#    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             #
############################################################################

import sys, os, traceback
import seiscomp3.Core, seiscomp3.Client, seiscomp3.IO

beforeP = afterP = 360

def usage(exitcode=0):
    sys.exit(exitcode)

class RecordDumperApp(seiscomp3.Client.Application):

    def __init__(self, argc, argv):
        seiscomp3.Client.Application.__init__(self, argc, argv)
        self.setMessagingEnabled(False)
        self.setDatabaseEnabled(True, True)
        self.setLoggingToStdErr(True)
        self.setDaemonEnabled(False)
        self.setRecordStreamEnabled(True)

    def createCommandLineDescription(self):
        try:
            try:
                self.commandline().addGroup("Dump")
                self.commandline().addStringOption("Dump", "event,E", "ID of event to dump")
#               self.commandline().addStringOption("Dump", "origin,O", "ID of origin to dump")

                self.commandline().addGroup("Input")
                self.commandline().addStringOption("Input", "format,f", "input format to use (xml [default], zxml (zipped xml), binary)")
                self.commandline().addStringOption("Input", "input,i", "input file, default: stdin")
            except:
                seiscomp3.Logging.warning("caught unexpected error %s" % sys.exc_info())
            return True
        except:
            info = traceback.format_exception(*sys.exc_info())
            for i in info: sys.stderr.write(i)
            sys.exit(-1)

    def validateParameters(self):
        try:
            if seiscomp3.Client.Application.validateParameters(self) == False:
                return False
            if not self.commandline().hasOption("event"):
                self.setDatabaseEnabled(False, False)
            return True
        except:
            info = traceback.format_exception(*sys.exc_info())
            for i in info: sys.stderr.write(i)
            sys.exit(-1)

    def _getDistancesArrivalsSorted(self, org):
        # sort arrival list by distance
        dist_arr = []
        for i in xrange(org.arrivalCount()):
            arr = org.arrival(i)
            if arr.weight() < 0.1:
                continue
            pha = arr.phase().code()
            if pha not in ["P","Pn","Pg","Pb","Pdiff","PKP"]:
                continue
            dist_arr.append((arr.distance(), arr))
        dist_arr.sort()
        return dist_arr

    def _getPicks(self, org):
        if self._dbq is None: return {}
        orid = org.publicID()
        pick = {}
        for obj in self._dbq.getPicks(orid):
            p = seiscomp3.DataModel.Pick.Cast(obj)
            key = p.publicID()
            pick[key] = p
        return pick

    def _getAmplitudes(self, org):
        if self._dbq is None: return {}
        orid = org.publicID()
        ampl = {}
        for obj in self._dbq.getAmplitudesForOrigin(orid):
            amp = seiscomp3.DataModel.Amplitude.Cast(obj)
            key = amp.publicID()
            ampl[key] = amp
        return ampl

    def _dumpWaveform(self, evt, org):
        waveform_windows = []
        for dist, arr in self.dist_arr:
            p = seiscomp3.DataModel.Pick.Find(arr.pickID())
            if p is None:
                continue
            wfid = p.waveformID()
            net = wfid.networkCode()
            sta = wfid.stationCode()
            loc = wfid.locationCode()
            cha = wfid.channelCode()

            t0 = p.time().value()
            t1, t2 = t0 + seiscomp3.Core.TimeSpan(-beforeP), t0 + seiscomp3.Core.TimeSpan(afterP)
            waveform_windows.append( (t1, t2, net, sta, loc, cha) )

        # request waveforms and dump them to one file per stream
        data = {}
        count = 1
        stream = seiscomp3.IO.RecordStream.Open(self.recordStreamURL())
        stream.setTimeout(300)
        for t1, t2, net, sta, loc, cha in waveform_windows:
            cha = cha[:2]+"?"
            stream.addStream(net, sta, loc, cha, t1, t2)
        input = seiscomp3.IO.RecordInput(stream, seiscomp3.Core.Array.INT, seiscomp3.Core.Record.SAVE_RAW)
        while 1:
            rec = input.next()
            if not rec: break
            sys.stderr.write("%-20s %6d\r" % (rec.streamID(), count))
            count += 1
            if not rec.streamID() in data:
                data[rec.streamID()] = ""

            # append binary (raw) MiniSEED record to data string
            data[rec.streamID()] += rec.raw().str()
        sys.stderr.write("Read %d records for %d streams\n" % (count, len(data.keys())))

        for streamID in data:
            file(evt.publicID()+"/"+streamID+".mseed", "w").write(data[streamID])

    def _dumpPickList(self, evt, org, filename="picks"):
        ampl = self._getAmplitudes(org)
        mags = {}
        nmag = org.stationMagnitudeCount()
        for i in xrange(nmag):
            mag = org.stationMagnitude(i)
            typ = mag.type()
            if typ == "MLv": typ = "ML"
            if typ not in mags:
                mags[typ] = {}

            sta = mag.waveformID().stationCode()
            mags[typ][sta] = mag

        pick_lines = []
        for dist, arr in self.dist_arr:
            p = seiscomp3.DataModel.Pick.Find(arr.pickID())
            if p is None:
                continue

            snr = 0.
            for key in ampl:
                amp = ampl[key]
                if amp.type() != "snr":
                    continue
                if amp.pickID() == arr.pickID():
                    snr = amp.amplitude().value()

            wfid = p.waveformID()
            net = wfid.networkCode()
            sta = wfid.stationCode()
            loc = wfid.locationCode()
            cha = wfid.channelCode()

            t0 = p.time().value()

            pst = "A"
            if p.evaluationMode() != seiscomp3.DataModel.AUTOMATIC:
                pst = "M"
            amp = per = 0.
            for typ in ["mb","ML","mB"]:
                try:
                    m = mags[typ][sta]
                    if typ=="mb":
                        ampid = m.amplitudeID()
                        a = seiscomp3.DataModel.Amplitude.Find(ampid)
                        if a is None:
                            obj = self._dbq.loadObject(seiscomp3.DataModel.Amplitude.TypeInfo(), ampid)
                            a   = seiscomp3.DataModel.Amplitude.Cast(obj)
                        if a:
                            per = a.period().value()
                            amp = a.amplitude().value()
                except KeyError:
                    pass

            line = "%s %-2s %-6s %-3s %-2s %6.1f %10.3f %4.1f %1s %-6s %s" \
                % (t0.toString("%Y-%m-%d %H:%M:%S.%f000000")[:22], net, sta, cha, [loc,"__"][loc==""], snr, amp, per, pst, arr.phase().code(), p.publicID())
            pick_lines.append(((dist,t0), line))

        # dump the pick lines to one single file
        pick_lines.sort()
        file(evt.publicID()+"/"+filename, "w").write("%s\n" % "\n".join([ line for x,line in pick_lines ]))

    def dumpEvent(self, eventID):
        self._dbq = seiscomp3.DataModel.DatabaseQuery(self.database())
        evt = self._dbq.loadObject(seiscomp3.DataModel.Event.TypeInfo(), eventID)
        evt = seiscomp3.DataModel.Event.Cast(evt)
        if evt is None:
            raise TypeError, "unknown event '" + eventID + "'"

        originID = evt.preferredOriginID()
        org = self._dbq.loadObject(seiscomp3.DataModel.Origin.TypeInfo(), originID)
        org = seiscomp3.DataModel.Origin.Cast(org)
        if not org:
            seiscomp3.Logging.error("origin '%s' not loaded" % originID)
            return

        # create event directory
        try:
            os.mkdir(evt.publicID())
        except OSError:
            pass # we interpret this as already existing event directory

        self.dist_arr = self._getDistancesArrivalsSorted(org)
        pick = self._getPicks(org)
        seiscomp3.Logging.info("dumping waveforms for preferred origin '%s'" % org.publicID())
        self._dumpWaveform(evt, org)
        seiscomp3.Logging.info("dumping picks for preferred origin '%s'" % org.publicID())
        self._dumpPickList(evt, org)

        #################################################################
        ### If you don't want to dump the picks for the largest automatic
        ### origin, then uncomment this:
        # return # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        #################################################################

        # find largest automatic origin

        # loop over all origins
        origins = [ org for org in self._dbq.getOrigins(eventID) ]
        # This DB query must be finished before we start iterating over the origins
        # as otherwise we get netsted queries, which are not allowed.

        sorted = []
        for org in origins:
            # there should be a cleaner way...
            org = seiscomp3.DataModel.Origin.Cast(org)
            if not org: continue
            org = self._dbq.loadObject(seiscomp3.DataModel.Origin.TypeInfo(), org.publicID())
            if not org: continue
            org = seiscomp3.DataModel.Origin.Cast(org)
            if not org: continue
            try:
                if org.evaluationMode() != seiscomp3.DataModel.AUTOMATIC:
                    continue
            except:
#               print org.publicID(), "failed"
                continue
#           print org.creationInfo().author(), org.creationInfo().agencyID()
            sorted.append( (org.arrivalCount(), org) )
        # get origin with largest arrival count
        sorted.sort()
        arrivalCount, org = sorted[-1]
        seiscomp3.Logging.info("dumping picks for automatic origin '%s'" % org.publicID())
        # we need to load arrivals and picks for that origin as well
        self.dist_arr = self._getDistancesArrivalsSorted(org)
        pick = self._getPicks(org)
        self._dumpPickList(evt, org, filename="picks-automatic")

    def run(self):
        try:
            try:
                eventID = self.commandline().optionString("event")
            except:
                sys.stderr.write("You must specify event id\n")
                return False

            try:
                self.dumpEvent(eventID)
            except Exception, exc:
                sys.stderr.write("ERROR: " + str(exc) + "\n")
                return False

            return True

        except:
            info = traceback.format_exception(*sys.exc_info())
            for i in info: sys.stderr.write(i)
            sys.exit(-1)

def main():
    app = RecordDumperApp(len(sys.argv), sys.argv)
    app()

if __name__ == "__main__":
    main()
